package ru.dbotthepony.kommons.util

import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.concurrent.withLock
import kotlin.reflect.KProperty0

interface ListenableDelegate<V> : Delegate<V>, Listenable<V> {
	class UnboundBox<V> : ListenableDelegate<V> {
		private var value: Any? = Companion
		private val listeners = Listenable.Impl<V>()

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return listeners.addListener(listener)
		}

		override fun get(): V {
			val get = value
			if (get === Companion) throw UninitializedPropertyAccessException()
			return get as V
		}

		fun clearListeners() {
			listeners.clear()
		}

		override fun accept(t: V) {
			value = t
		}

		private companion object
	}

	class Box<V>(private var value: V) : ListenableDelegate<V> {
		private val listeners = Listenable.Impl<V>()

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return listeners.addListener(listener)
		}

		fun clearListeners() {
			listeners.clear()
		}

		override fun get(): V {
			return value
		}

		override fun accept(t: V) {
			if (value != t) {
				value = t
				listeners.accept(t)
			}
		}
	}

	class SmartBox<V>(value: V, private val getter: DelegateGetter<V> = DelegateGetter.passthrough(), private val setter: DelegateSetter<V> = DelegateSetter.passthrough()) : ListenableDelegate<V> {
		private val value = Box(value)

		fun clearListeners() {
			value.clearListeners()
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return value.addListener(listener)
		}

		override fun get(): V {
			return getter(value)
		}

		override fun accept(t: V) {
			setter(value, t)
		}
	}

	abstract class AbstractShadow<V>(private val parent: Supplier<V>) : ListenableDelegate<V>, ValueObserver<V> {
		@Volatile
		private var shadow: Any? = Companion
		@Volatile
		private var observed: Any? = Companion
		private val listeners = Listenable.Impl<V>()
		private val lock = ReentrantLock()

		protected abstract fun compare(a: V, b: V): Boolean
		protected abstract fun copy(value: V): V

		override fun get(): V {
			if (shadow !== Companion) {
				return shadow as V
			}

			return parent.get()
		}

		override fun accept(t: V) {
			val shadow = shadow
			this.shadow = t

			if (shadow != t) {
				listeners.accept(t)
			}
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return listeners.addListener(listener)
		}

		fun clearListeners() {
			listeners.clear()
		}

		override fun observe(): Boolean {
			if (shadow !== Companion)
				return false

			val get = parent.get()

			if (observed === Companion || !compare(observed as V, get)) {
				lock.withLock {
					if (shadow !== Companion)
						return false

					if (observed === Companion || !compare(observed as V, get)) {
						observed = copy(get)
						listeners.accept(get)
						return true
					}
				}
			}

			return false
		}

		override fun getAndObserve(): Pair<V, Boolean> {
			if (shadow !== Companion)
				return (shadow as V) to false

			val get = parent.get()

			if (observed === Companion || !compare(observed as V, get)) {
				lock.withLock {
					if (shadow !== Companion)
						return (shadow as V) to false

					if (observed === Companion || !compare(observed as V, get)) {
						observed = copy(get)
						listeners.accept(get)
						return get to true
					}
				}
			}

			return get to false
		}

		private companion object
	}

	class Shadow<V>(parent: Supplier<V>) : AbstractShadow<V>(parent) {
		constructor(parent: KProperty0<V>) : this(parent::get)

		override fun compare(a: V, b: V): Boolean {
			return a == b
		}

		override fun copy(value: V): V {
			return value
		}
	}

	class CustomShadow<V>(parent: Supplier<V>, private val comparator: (V, V) -> Boolean, private val copy: (V) -> V) : AbstractShadow<V>(parent) {
		constructor(parent: KProperty0<V>, comparator: (V, V) -> Boolean, copy: (V) -> V) : this(parent::get, comparator, copy)

		override fun compare(a: V, b: V): Boolean {
			return comparator(a, b)
		}

		override fun copy(value: V): V {
			return copy.invoke(value)
		}
	}

	companion object {
		/**
		 * Returns smart box if [getter] or [setter] is not passthrough
		 */
		@JvmStatic
		fun <V> maskSmart(value: V, getter: DelegateGetter<V>, setter: DelegateSetter<V>): ListenableDelegate<V> {
			if (getter === DelegateGetter.passthrough<V>() && setter === DelegateSetter.passthrough<V>()) {
				return Box(value)
			} else {
				return SmartBox(value, getter, setter)
			}
		}
	}
}
